unit QueryThreads;

interface

uses Windows, Winsock2, SysUtils, SyncObjs;

type
  ThreadHandelrProc = function (param: pointer): dword; stdcall;

  PQueryThreadParam = ^TQueryThreadParam;
  TQueryThreadParam = record
    done: boolean;
    value: TBytes;
    //realm: TRealm;
    socket: TSocket;
    query_id: dword;
    custom: TBytes;
  end;

  TQueryId = class
  private
    FCS: TCriticalSection;
    FId: dword;
    FIdArray: array[0..MAXWORD]of PQueryThreadParam;
    function GetNewParam: PQueryThreadParam;
    function GetValue(idx: dword): PQueryThreadParam;
    procedure SetValue(idx: dword; value: PQueryThreadParam);
  public
    constructor Create;
    destructor Destroy;override;
    procedure Remove(idx: dword);
    procedure Lock;
    procedure Unlock;
    property NewParam: PQueryThreadParam read GetNewParam;
    property Value[idx: dword]: PQueryThreadParam read GetValue write SetValue;default;
  end;

  TQueryThread = class
  private
    FThreadParams: array of PQueryThreadParam;
    FParamCount: dword;
    FQueryCmd: byte;
    FThreadId: dword; //array of
    FThreadHandle: THandle; //TWOHandleArray;
    FProc: ThreadHandelrProc;
    //function GetParam(idx: integer): PQueryThreadParam;
    //procedure SetParam(idx: integer; p: PQueryThreadParam);
    function GetParamByIndex(idx: dword): PQueryThreadParam;
  public
    signal_terminate: boolean;
    constructor Create(count: dword; proc: ThreadHandelrProc; QueryCmd: byte);
    destructor Destroy;override;
    procedure StartThread;
    procedure DoTerminate;
    function Add(idx: dword): PQueryThreadParam;
    property ParamByIndex[idx: dword]: PQueryThreadParam read GetParamByIndex;
    //property PParam[idx: integer]: PQueryThreadParam read GetParam write SetParam; default;
    property ParamCount: dword read FParamCount;
  end;

  function GetCharsCountThread(param: pointer): dword; stdcall;

implementation

uses RealmsHandler, LogHandler;

{ TQueryId }

constructor TQueryId.Create;
begin
    FCS:= TCriticalSection.Create;
    ZeroMemory(@FIdArray, sizeof(FIdArray));
    FId:= 0;
end;

destructor TQueryId.Destroy;
begin
    FCS.Free;
    inherited;
end;

function TQueryId.GetNewParam: PQueryThreadParam;
var
  i: dword;
begin
    FCS.Enter;
    for i := low(FIdArray) to high(FIdArray) do begin
      if FIdArray[i] = nil then
        break
      else begin
        inc(FId);
        FId:= FId and $FFFF;
      end;
    end;
    GetMem(FIdArray[FId], sizeof(TQueryThreadParam));
    ZeroMemory(FIdArray[FId], sizeof(TQueryThreadParam));
    result:= FIdArray[FId];
    result.query_id:= FId;
    result.done:= false;
    FCS.Leave;
end;

procedure TQueryId.Remove(idx: dword);
var
  id: dword;
begin
    FCS.Enter;
    id:= idx and $FFFF;
    FIdArray[id].value:= nil;
    FIdArray[id].custom:= nil;
    FreeMem(FIdArray[id]);
    FIdArray[id]:= nil;
    FCS.Leave;
end;

function TQueryId.GetValue(idx: dword): PQueryThreadParam;
begin
    result:= FIdArray[idx];
end;

procedure TQueryId.SetValue(idx: dword; value: PQueryThreadParam);
begin
    FIdArray[idx]:= value;
end;

procedure TQueryId.Lock;
begin
    FCS.Enter;
end;

procedure TQueryId.Unlock;
begin
    FCS.Leave;
end;

{ TThreadParamHandlers }

function GetCharsCountThread(param: pointer): dword; stdcall;
var
  tp: TQueryThread;
  i: integer;
  done: boolean;
begin
    result:= 0;
    tp:= TQueryThread(param);
    while not(tp.signal_terminate) do begin 
      //break on param^.done changed
      done:= true;
      for i := 0 to tp.ParamCount - 1 do begin
        if not(tp.ParamByIndex[i].done) then begin //QueryIds[tp.PParam[i].query_id].done
          done:= false;
          break;
        end;
      end;
      if done then break;
    end;
end;

{ TQueryThread }

constructor TQueryThread.Create(count: dword; proc: ThreadHandelrProc; QueryCmd: byte);
begin
    // ,           done  terminate
    signal_terminate:= false;
    FQueryCmd:= QueryCmd;
    FParamCount:= count;
    setlength(FThreadParams, FParamCount);

    @FProc:= @proc;

    FThreadHandle:= INVALID_HANDLE_VALUE;
end;

function TQueryThread.Add(idx: dword): PQueryThreadParam;
begin
    result:= QueryIds.NewParam;
    FThreadParams[idx]:= result;
end;

procedure TQueryThread.DoTerminate;
var
  wait_res: dword;
begin
    if (FThreadHandle = INVALID_HANDLE_VALUE) then exit;

    wait_res:= WaitForMultipleObjects(1, @FThreadHandle, true, 5000); //max len = MAXIMUM_WAIT_OBJECTS
    if (wait_res = WAIT_FAILED) then begin
      AddToLog('WaitForMultipleObjects failed');
      ErrorCheck;
    end;
    if (wait_res = WAIT_TIMEOUT) then begin
      signal_terminate:= true;
      wait_res:= WaitForMultipleObjects(1, @FThreadHandle, true, 100); 
      if (wait_res = WAIT_FAILED) then begin
        AddToLog('WaitForMultipleObjects failed');
        ErrorCheck;
      end;
    end;

    CloseHandle(FThreadHandle);
    FThreadHandle:= INVALID_HANDLE_VALUE;
end;

destructor TQueryThread.Destroy;
var
  i: integer;
begin
    DoTerminate;

    for i := 0 to FParamCount - 1 do
      QueryIds.Remove(FThreadParams[i].query_id);

    FThreadParams:= nil;
    inherited;
end;

{function TQueryThread.GetParam(idx: integer): PQueryThreadParam;
begin
    result:= FThreadParams[idx];
end;

procedure TQueryThread.SetParam(idx: integer; p: PQueryThreadParam);
begin
    FThreadParams[idx]:= p;
end;}

function TQueryThread.GetParamByIndex(idx: dword): PQueryThreadParam;
begin
    result:= FThreadParams[idx];
end;

procedure TQueryThread.StartThread;
var
   i: integer;
   hThread: dword;
begin
    hThread:= CreateThread(nil, 0, @FProc, self, 0, FThreadId);
    if hThread = 0 then begin
       AddToLog('CreateThread failed to create worker thread');
       ErrorCheck;
       exit;
    end;
		FThreadHandle:= hThread;
    for i := 0 to FParamCount - 1 do
      ExecThreadQuery(FThreadParams[i].socket, FThreadParams[i].custom, FQueryCmd);
end;

end.
