unit RealmsHandler;

interface

uses Windows, SysUtils, WinSock2, Classes, ComCtrls, SyncObjs,
  IOCPHandler, TypesDefs, PostgresClasses, libpq_fe, QueryThreads;

type
  RealmProcs = function(pdata: pointer; count: dword; sckt: TSocket):dword;

  TRealm = class
  public
    name: UTF8String;
    address_port: string;
    //db_name, db_user, db_password: string;
    population: single;
    online,
    r_type,     //0 - normal, 1 - pvp, 6 - rp, 8 - rppvp
    locked,
    color,      //0 - green 1 - red 2 - gray (offline)
    locale,     
    unk_6: byte;
    socket: TSocket;
    listItem: TListItem;
    //irPQ: IPostgres;
    constructor Create(sock: TSocket);
    destructor Destroy;override;
  end;

  procedure OnReadRealmSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
  procedure OnWriteRealmSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
  procedure OnAcceptRealmSocket(ppsContext: PPER_SOCKET_CONTEXT);
  procedure OnDisconnectRealmSocket(ppsContext: PPER_SOCKET_CONTEXT);

  procedure ExecThreadQuery(socket: TSocket; data: TBytes; QueryCmd: byte);

const
    R_REALM_SEND_ID_HASH            = $00;
    R_REALM_GET_POPULATION          = $01;
    R_REALM_GET_CHARSCOUNT          = $02;

    WRONG_ID                        = $FFFFFFFF;

var
  QueryIds: TQueryId;

implementation

uses UnitMain, ByteArrayHandler, LogHandler, LoginHandler;

const
    L_REALM_REGISTER                = $00;
    L_REALM_GET_ID_HASH             = $01;
    L_REALM_CHARSCOUNT              = $02;
    L_REALM_POPULATION              = $03;

    LI_ADDR_PORT_IDX                = 0;
    LI_ONLINE_IDX                   = 1;
    LI_POP_IDX                      = 2;
    LI_RTYPE_IDX                    = 3;
    LI_LOCKED_IDX                   = 4;
    LI_COLOR_IDX                    = 5;
    LI_LOCALE_IDX                   = 6;
    LI_UNK_IDX                      = 7;

    AUTH_PACKET_DATAOFFSET          = 3;

var
  ReadBufBySock: array[0..MAXWORD]of TReadBuffer;
  SendBufBySock: array[0..MAXWORD]of TSendBuffer;
  RealmBySock: array[0..MAXWORD]of TRealm;
  RealmProcTable : array[L_REALM_REGISTER..L_REALM_POPULATION] of RealmProcs;

{ TRealm }

constructor TRealm.Create(sock: TSocket);
begin
    RealmBySock[sock]:= self;
end;

destructor TRealm.Destroy;
begin
    RealmBySock[socket]:= nil;
    inherited;
end;

{ handlers }

procedure OnAcceptRealmSocket(ppsContext: PPER_SOCKET_CONTEXT);
begin
    //ThreadedAccountsBySockHT.Remove(ppsContext.Socket);
    ReadBufBySock[ppsContext.Socket]:= TReadBuffer.Create;
    ppsContext.pIOReadContext.wsa_buf.len:= ReadBufBySock[ppsContext.Socket].GetBuffFreeSpace;
    ppsContext.pIOReadContext.wsa_buf.buf:= ReadBufBySock[ppsContext.Socket].GetDataPtr;
    SendBufBySock[ppsContext.Socket]:= TSendBuffer.Create(SIZE1K);
end;

procedure OnDisconnectRealmSocket(ppsContext: PPER_SOCKET_CONTEXT);
begin
    //ThreadedAccountsBySockHT.Remove(ppsContext.Socket);
    FreeAndNil(ReadBufBySock[ppsContext.Socket]);
    FreeAndNil(SendBufBySock[ppsContext.Socket]);

    if RealmBySock[ppsContext.Socket] <> nil then begin
      RealmBySock[ppsContext.Socket].online:= 0;
      RealmBySock[ppsContext.Socket].listItem.SubItems[LI_ONLINE_IDX]:= 'false';
      RealmBySock[ppsContext.Socket]:= nil;
    end;
end;

procedure OnReadRealmSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
var
   cmd: byte;
   tempRLB: TReadBuffer;
   data: pointer;
   done: dword;
begin
     if ppsContext.Socket > high(ReadBufBySock) then exit;
     tempRLB:= ReadBufBySock[ppsContext.Socket];
     if tempRLB = nil then exit;
     if not(tempRLB.IncReceived(count)) then begin
        RealmsSock.CloseContextSocket(ppsContext.Socket, ClientIoRead);
        exit;
     end;

     {lmsg:='';lstr:='';
     for i:=0 to pred(received_len) do lmsg:=lmsg+IntToHex(Buf[i],2)+' ';
     for i:=0 to pred(received_len) do if Buf[i]<>0 then lstr:=lstr+char(Buf[i])+' ';
     AddToLog('receive auth socket ='+IntToStr(Sock));
     AddToLog('Msg='+lmsg+' ('+lstr+')'+#$d#$a);}

     data:= tempRLB.GetDataPtr;
     cmd:= TBytes(data)[0]; //ppsContext.pIOReadContext.Buffer[0];
     AddToLog('receive auth opcode = ' + IntToStr(cmd));
     done:= tempRLB.GetDataCount;
     if (cmd < length(RealmProcTable))and(@RealmProcTable[cmd] <> nil) then begin
        done:= RealmProcTable[cmd](data, tempRLB.GetDataCount, ppsContext.Socket);
     end;
     if ReadBufBySock[ppsContext.Socket] = nil then exit; //dced during handle
     if done > 0 then tempRLB.ReleaseBufferCount(done);
     ppsContext.pIOReadContext.wsa_buf.len:= tempRLB.GetBuffFreeSpace;
     ppsContext.pIOReadContext.wsa_buf.buf:= tempRLB.GetDataPtr;
end;

procedure OnWriteRealmSocket(ppsContext: PPER_SOCKET_CONTEXT; count: dword);
begin
     Assert(SendBufBySock[ppsContext.Socket] <> nil,'SendBufBySock[ppsContext.Socket] is nil');
     SendBufBySock[ppsContext.Socket].IncSent(count);
     SendBufBySock[ppsContext.Socket].Send(RealmsSock, ppsContext.Socket);
end;

{ responses to RealmServer requests/commands }

function RRecvRegister(pdata: pointer; count: dword; sckt: TSocket):dword;
type
  PREG_PACKET = ^TREG_PACKET;
  TREG_PACKET = packed record
    OpCode: Byte;
    Size: word;
    population: single;
    online,
    r_type,     //0 - normal, 1 - pvp
    locked,
    color,      //0 - green 1 - red 2 - gray (offline)
    locale,
    unk_6: byte;
  end;
var
  i, idx, offs: integer;
  list: TStringList;
  name: UTF8String;
  addr_port: string;
  p: PREG_PACKET;
  r: TRealm;
const
  LOCALES: array[0..17]of string = ('hidden', 'Development', 'United States', 'Oceanic', 'Latin America',
        'Tournament', 'Korea', 'Tournament', 'English', 'German', 'French', 'Spanish',
        'Russian (ruRU only)', 'Tournament', 'Taiwan', 'Tournament', 'China', 'CN1');
begin
    result:= 0;
    if count < AUTH_PACKET_DATAOFFSET then exit;

    p:= pdata;

    if count < p.Size + AUTH_PACKET_DATAOFFSET then exit;
    result:= p.Size + AUTH_PACKET_DATAOFFSET;

    offs:= sizeof(TREG_PACKET);
    name:= PChar(@TBytes(pdata)[offs]);
    inc(offs, length(name) + 1);
    addr_port:= PChar(@TBytes(pdata)[offs]);

    list:= RealmsList.Lock;
    try
      idx:= list.IndexOf(name);
      if idx >= 0 then
        r:= TRealm(list.Objects[idx])
      else begin
        r:= TRealm.Create(sckt);
        list.AddObject(name, r);
        r.listItem:= Form1.ListViewRealms.Items.Insert(Form1.ListViewRealms.Items.Count);
        r.listItem.Caption:= name;
        for i:= 0 to 8 do r.listItem.SubItems.Add('');
      end;

      r.name:= name;
      r.address_port:= addr_port;
      r.online:= p.online;
      r.population:= p.population;
      r.r_type:= p.r_type;
      r.locked:= p.locked;          
      r.color:= p.color;
      r.locale:= p.locale;          
      r.unk_6:= p.unk_6;
      r.socket:= sckt;
      RealmBySock[sckt]:= r;

      r.listItem.SubItems[LI_ADDR_PORT_IDX]:= addr_port;
      if r.online <> 0 then
        r.listItem.SubItems[LI_ONLINE_IDX]:= 'true'
      else r.listItem.SubItems[LI_ONLINE_IDX]:= 'false';
      r.listItem.SubItems[LI_POP_IDX]:= FloatToStr(r.population);
      case r.r_type of
        0: r.listItem.SubItems[LI_RTYPE_IDX]:= 'normal';
        1: r.listItem.SubItems[LI_RTYPE_IDX]:= 'pvp';
        6: r.listItem.SubItems[LI_RTYPE_IDX]:= 'rp';
        8: r.listItem.SubItems[LI_RTYPE_IDX]:= 'rppvp';
      else r.listItem.SubItems[LI_RTYPE_IDX]:= IntToStr(r.r_type);
      end;
      if r.locked <> 0 then
        r.listItem.SubItems[LI_LOCKED_IDX]:= 'true'
      else r.listItem.SubItems[LI_LOCKED_IDX]:= 'false';
      r.listItem.SubItems[LI_COLOR_IDX]:= IntToStr(r.color);
      if (r.locale <= high(LOCALES)) then
        r.listItem.SubItems[LI_LOCALE_IDX]:= LOCALES[r.locale]
      else r.listItem.SubItems[LI_LOCALE_IDX]:= IntToStr(r.locale);
      r.listItem.SubItems[LI_UNK_IDX]:= IntToStr(r.unk_6);
    finally
      RealmsList.UnLock;
    end;
end;

function RSendIdHash(pdata: pointer; count: dword; sckt: TSocket):dword;
type
  PGETIDHASH_PACKET = ^TGETIDHASH_PACKET;
  TGETIDHASH_PACKET = packed record
    OpCode: Byte;
    Size: word;
    query_id: dword;
    name: char;
  end;
var
  pkt: PGETIDHASH_PACKET;
  account_name: UTF8String;
  SS: TSendSession;
  wrong_hash: THashArray;
  CurrAuthPlayer: TThreadedAuthPlayer;
  ap: TAuthPlayer;
begin
    result:= 0;
    if count < AUTH_PACKET_DATAOFFSET then exit;

    pkt:= pdata;

    if count < pkt.Size + AUTH_PACKET_DATAOFFSET then exit;
    result:= pkt.Size + AUTH_PACKET_DATAOFFSET;

    account_name:= PChar(@pkt.name);

    CurrAuthPlayer:= ThreadedAccountsByNameHT[account_name];
    SS:= TSendSession.Create(SIZE128);
    SS.InitAuth(R_REALM_SEND_ID_HASH);
    SS.Add(word(SIZE1K));  //packet size
    SS.Add(pkt.query_id);
    if CurrAuthPlayer = nil then begin
      AddToLog('account retrieve error: ' + account_name);
      SS.Add(dword(WRONG_ID));
      SS.Add(@wrong_hash, sizeof(THashArray));
    end
    else begin
      ap:= CurrAuthPlayer.Lock;
      try
        SS.Add(ap.id);
        SS.Add(@ap.Hash, sizeof(THashArray));
      finally
        CurrAuthPlayer.UnLock;
      end;
    end;
    SS.PutToPos(word(SS.Offs - AUTH_PACKET_DATAOFFSET), 1); //store packet length
    if SendBufBySock[sckt] <> nil then begin
      SendBufBySock[sckt].AddData(SS.data, SS.Offs);
      SendBufBySock[sckt].Send(RealmsSock, sckt);
    end;
    SS.Free;
end;

function RRecvCharsCount(pdata: pointer; count: dword; sckt: TSocket):dword;
type
  PCHARSCOUNT_PACKET = ^TCHARSCOUNT_PACKET;
  TCHARSCOUNT_PACKET = packed record
    OpCode: Byte;
    Size: word;
    query_id: dword;
    account_id: dword;
    count: dword;
  end;
var
  p: PCHARSCOUNT_PACKET;
begin
    result:= 0;
    if count < AUTH_PACKET_DATAOFFSET then exit;

    p:= pdata;

    if count < p.Size + AUTH_PACKET_DATAOFFSET then exit;
    result:= p.Size + AUTH_PACKET_DATAOFFSET;

    if QueryIds.Value[p.query_id] = nil then exit; //too late

    QueryIds.Lock;
    try
      setlength(QueryIds.Value[p.query_id].value, sizeof(dword));
      move(p.count, QueryIds.Value[p.query_id].value[0], sizeof(dword));
      QueryIds.Value[p.query_id].done:= true;
    finally
      QueryIds.Unlock;
    end;
end;

procedure ExecThreadQuery(socket: TSocket; data: TBytes; QueryCmd: byte);
var
  SS: TSendSession;
  len: dword;
begin
    len:= SIZE32 + length(data);
    SS:= TSendSession.Create(len);
    SS.InitAuth(QueryCmd);
    SS.Add(word(len));  //packet size
    SS.Add(data, length(data));
    SS.PutToPos(word(SS.Offs - 3), 1); //store packet length
    if SendBufBySock[socket] <> nil then begin
      SendBufBySock[socket].AddData(SS.data, SS.Offs);
      SendBufBySock[socket].Send(RealmsSock, socket);
    end;
    SS.Free;
end;

procedure InitHandlers;
var
  i: integer;
begin
   QueryIds:= TQueryId.Create;

   ZeroMemory(@ReadBufBySock[0], sizeof(ReadBufBySock));
   ZeroMemory(@SendBufBySock[0], sizeof(SendBufBySock));
   ZeroMemory(@RealmBySock[0], sizeof(RealmBySock));

   for i:= low(RealmProcTable) to high(RealmProcTable) do
       @RealmProcTable[i]:= nil;

   @RealmProcTable[L_REALM_REGISTER]        := @RRecvRegister;
   @RealmProcTable[L_REALM_GET_ID_HASH]     := @RSendIdHash;
   @RealmProcTable[L_REALM_CHARSCOUNT]      := @RRecvCharsCount;
   //@RealmProcTable[L_REALM_POPULATION]     := @RPopulation;

end;

procedure ReleaseBuffers;
var
  i: integer;
begin
    for i:=0 to high(ReadBufBySock) do
      if ReadBufBySock[i] <> nil then
        ReadBufBySock[i].Free;
    for i:=0 to high(SendBufBySock) do
      if SendBufBySock[i] <> nil then
        SendBufBySock[i].Free;

    QueryIds.Free;
end;

initialization

  InitHandlers;

finalization

  ReleaseBuffers;

end.
